# __ File: executionControl.py ______________________________________________________________
# Implements : class executionControl
# Author     : Juan de Lara
# Description: Window with some widgets to control the execution of the
#              graph grammars if STEP by STEP mode has been selected.
#              Initially generated with ATOM3, added some widgets by hand.
# Modified   :
#   - 21 Oct 2001. Header added
# ___________________________________________________________________________________________

from Tkinter import *

class executionControl(Frame):

    def __init__(self, grs):
       "Constructor, takes the frame's parent and the GraphRewritingSys instance as arguments"
       self.rootWindow = Tk()                                                                                 # create a new window...
       Frame.__init__(self, self.rootWindow)                             # call the Frame constructor
       self.graphRWS = grs                                      # Store the Graph Rewriting System
       self.finished = 0                                                # indicates if the GG have finished
       self.rootWindow.title("Graph-Grammar execution controls")    # set window's title
       ggpanel = Frame(self.rootWindow, relief = RAISED)
       Label(ggpanel,text="Executing Graph-Grammar: ").grid(row=1, col=0, sticky=W)
       self.currGG = Label(ggpanel,text=self.graphRWS.getCurrentGG())
       self.currGG.grid(row=1, col=1, sticky=W)

       Label(ggpanel,text="Last executed rule: ").grid(row=2, col=0, sticky=W)
       self.curRule = Label(ggpanel,text="")
       self.curRule.grid(row=2, col=1, sticky=W)

       self.stepButton = Button(ggpanel, text="Step", command = self.doStep)
       self.stepButton.grid(row=3, col=0, sticky=W)

       self.contButton = Button(ggpanel, text="Continuous", command = self.continuous)
       self.contButton.grid(row=3, col=2, sticky=W)

       self.closeButton = Button(ggpanel, text="Close", command = self.closeWindow)
       self.closeButton.grid(row=3, col=3, sticky=W)

       ggpanel.grid(row=0, col=0)

    def continuous(self):
       "Steps in the GG's execution until it finishes"
       legg = self.graphRWS.lastExecutedGG  #epp
       ggs  = self.graphRWS.GraphGrammars   #epp
       while not (legg >= len(ggs)):        #epp
          self.finished = self.graphRWS.doStep()
          self.curRule.config(text=self.graphRWS.getLastRule())
          if self.finished:
             self.currGG.config(text="Execution finished!")
          else:
             self.currGG.config(text=self.graphRWS.getCurrentGG())
          legg = self.graphRWS.lastExecutedGG  #epp
          ggs  = self.graphRWS.GraphGrammars   #epp
       self.graphRWS.parent.closeUnusedMetaModels()               # Modified 09 Sep 2002 by JL

    def doStep(self):
       "makes one step in the GG's execution"
       legg = self.graphRWS.lastExecutedGG  #epp
       ggs  = self.graphRWS.GraphGrammars   #epp
       done = (legg >= len(ggs))            #epp
       if not done:            # if we have not finished yet
          self.finished = self.graphRWS.doStep()
          self.curRule.config(text=self.graphRWS.getLastRule())
          if self.finished:
             self.currGG.config(text="Execution finished!")
             self.graphRWS.parent.closeUnusedMetaModels()               # Modified 09 Sep 2002 by JL
          else:
             self.currGG.config(text=self.graphRWS.getCurrentGG())

        # self.graphRWS.parent.closeUnusedMetaModels()                  # Modified 09 Sep 2002 by JL

    def closeWindow(self):
       self.rootWindow.withdraw()
       self.graphRWS.parent.closeUnusedMetaModels()


