#
# PNlink.py
#  Módulo que conecta las Redes de Petri con el grafo
#  de alcanzabilidad / cubrimiento
#  
#  Desde el botón correspondiente del metamodelo de Red de Petri
#  (PetriNets.py), se incluirá este módulo que generará el
#  modelo en formalismo ReachGraph
#
# Autor: Jaime López
# Ultima modificación : 2003208
# $Id$
#

from Tkinter import *
import debug
import ReachGraph
import ReachGraph_MM
import ASG_ReachGraph
from ASG import *
from ATOM3 import *

def OpenReachGraph (AT3Dialog, ATOM3Instance, SemanticObject) :
    #ATOM3Instance.graph = AT3Dialog.parent.ASGroot
    l = PNlink()
    #print "AT3Dialog.parent = ", AT3Dialog.parent.ASGroot , "ATOM3Instance = ", ATOM3Instance.ASGroot
    currentState = l.pn_getCurrentState( AT3Dialog.parent.ASGroot )
    l.convert(AT3Dialog.parent, currentState, None, ATOM3Instance)
    for state in ATOM3Instance.ASGroot.listNodes["RGnode"]:
	    state.state.setValue(state.symName)

class PNlink :

	# Constructor
	def __init__(self) :
		self.iDebug = 0			# Imprimir informacin de debug
		self.iCubrimiento = 0		# 0=solo alcanzabilidad, 1=alcanzabilidad/cubrimiento
		self.GRID_SIZEX=230		# Tamaño de celda para posicionar nodos
		self.GRID_SIZEY=80		# Tamaño de celda para posicionar nodos
		self.iRec = 0			# Nivel de recursion actual

	#
	# Funciones de utilidad
	#
	
	# A partir de un diccionario, crea una lista
	# ordenada con sus valores
	# Permite ordenar por llave (por defecto) o por valor
	def _sort_dict(self, d, by_vals=0 ) :
		if by_vals : 
			items = [(v) for k,v in d.items()]
			items.sort()
		else :
			keys = d.keys()
			keys.sort()
			items = [ (d[k]) for k in keys ]
		return items
	
	# Convierte una lista en cadena
	def _list2str( self, l, sep = " ", print_brackets = 1 ) :
		s = ""
		if print_brackets : s = "[" + sep
		for i in l :
			s = s + str(i) + sep
		if print_brackets : s = s + "]"
		return s
	
	# Imprime con un cierto nivel de indentación
	def _print_indent ( self, s  ) :
		if self.iDebug == 1 :
			st = "%" + str(self.iRec) + "s " + s
			print st % ">"

	#
	# Funciones de manejo de la Red de Petri
	#
	
	# Obtiene el estado actual, compuesto
	# por los valores de los tokens. Ej: [2 1 0]
	def pn_getCurrentState ( self, pn ) :
		aPlaces = pn.listNodes.get('PNPlace')
	 	aState = {} 
		for i in aPlaces :
			aState[ i.name.getValue() ] = i.tokens.getValue()
		return aState
	
	# Obtiene la lista de estados
	# accesibles desde el estado actual
	def pn_getCandidates ( self, pn, aState ) :
		
		# Obtener la lista de transiciones habilitadas
		aHabTrans = []
		aTrans = pn.listNodes.get('PNTransition')
		for oTrans in aTrans :
		 	bHabilitada = 1
			for t in oTrans.in_connections_ :
				n  = t.in_connections_[0]
				if aState[ n.name.getValue() ] < t.weight.getValue() :
					bHabilitada = 0
					break	
			if bHabilitada : 
				aHabTrans.append( oTrans )
	
		# Generar los nuevos estados posibles (candidatos)
		# a partir de las transiciones habilitadas
		aCandidates = []
		for oTrans in aHabTrans :
			aNewState = aState.copy() 
			# Eliminar tokens en las entradas
			for t in oTrans.in_connections_ :
				n = t.in_connections_[0]
				if aNewState[ n.name.getValue() ] == "w" :
				  	aNewState[ n.name.getValue() ] = "w"
				else :
					aNewState[ n.name.getValue() ] = aNewState[ n.name.getValue() ] - t.weight.getValue()
			# Añadir tokens en las salidas
			for t in oTrans.out_connections_ :
				m = t.out_connections_[0]
				if aNewState[ m.name.getValue() ] == "w" :
				  	aNewState[ m.name.getValue() ] = "w"
				else :
					aNewState[ m.name.getValue() ] = aNewState[ m.name.getValue() ] + t.weight.getValue()
			# Añadir el nuevo estado posible
			aCandidates.append ( (aNewState, oTrans) )
		return aCandidates
	
	#
	# Funciones de manipulación del grafo de alcanzabilidad
	#
	def rg_addNode ( self, atom_3, aCurrentState, oParentNode = None, iPos=1, transition = None ) :
		sNode = self._list2str( self._sort_dict( aCurrentState ) )
		n = atom_3.createNewRGnode( atom_3, 0, 0 )
		n.__aCurrentState__ = aCurrentState
		n.state.setValue( sNode )
		
		symName = "["
		first = 1
		for name in aCurrentState.keys():
			if aCurrentState[name] == 1:
				if not first: symName = symName + ", "
				symName = symName + name
				first = 0
			elif aCurrentState[name] != 0:
				if not first: symName = symName + ", "				
				symName = symName + name + "("+str(aCurrentState[name])+")"
				first = 0
		symName = symName + "]"
		n.symName = symName
		n.graphObject_.ModifyAttribute("state", symName)
		n.graphObject_.Move(iPos*self.GRID_SIZEX, self.iRec*self.GRID_SIZEY)
	
		self._print_indent( "Aadiendo " + sNode )
		if oParentNode != None :
		 	self.rg_connectPair( atom_3, oParentNode, n, transition )	

		return n
	
	# Conecta dos nodos entre si
	def rg_connectPair ( self, atom_3, oParentNode, oChildNode, transition ) :
	 	#c = atom_3.createNewRGtrans( atom_3, 0, 0)
		#c.out_connections_.append(oChildNode)
		#c.in_connections_.append(oParentNode)
		#oChildNode.in_connections_.append(c)
		#oParentNode.out_connections_.append(c)
		self._print_indent( "Conectando "+oParentNode.state.getValue()+" con "+oChildNode.state.getValue() )
		atom_3.drawConnections((oParentNode, oChildNode))
		atom_3.newCreatedObject.transition.setValue(transition.name.getValue())
		atom_3.newCreatedObject.graphObject_.ModifyAttribute("transition", transition.name.getValue())


	# Determina si un nodo domina a otro (aState a oNode)
	def rg_isDominant( self, oNode, aState ) :
		iDomPos = []
		i = 0
		_aCurrentState = self._sort_dict( oNode.__aCurrentState__ ) 
		_aState = self._sort_dict( aState ) 
		for v in _aCurrentState :
			if _aState[i] == "w" and v!= "w": iDomPos = iDomPos+[i]
			else :
				if v == "w" and _aState[i] != "w": return []
				elif v != "w" and _aState[i] != "w":
					if int(_aState[i]) > int(v) : iDomPos = iDomPos+[i]
					elif int(_aState[i]) < int(v): return []
			i = i + 1
		# print _aState," dominates ",_aCurrentState," in position ",str(iDomPos)
		return iDomPos 

	# Renombra el estado de un nodo dominante en las posiciones de la lista iList
	# En su lugar, coloca "w"
	def rg_renameDominantNode( self, aState, iList ) :
		sState = self._list2str( self._sort_dict(aState) )
		name = aState.keys()
		name.sort()
		for pos in iList:
		   aState[name[pos]] = "w"
		s = self._list2str( self._sort_dict(aState) )
		# print "Renombrando " + sState + " => " + s 

	# Determina si hay un determinado estado en el grafo
	def rg_findState ( self, rg, state ) :
		s = self._list2str( self._sort_dict(state) )
	
		aNodes = rg.listNodes.get('RGnode')
		for n in aNodes :
			# Caso trivial : los nodos son identicos
			self._print_indent("Comparando "+s+" con "+n.state.getValue())
			if n.state.getValue() == s :
				self._print_indent("IGUALES!")
				return n
			else :
				self._print_indent("DISTINTOS!")	

		# Considerar dominancia de nodos
		if self.iCubrimiento :
			for n in aNodes :
				self._print_indent("Comparando dominancia "+s+" con "+n.state.getValue())
				iDomPos = self.rg_isDominant( n, state )
				if iDomPos != [] :
					self.rg_renameDominantNode( state, iDomPos )
					break
				else :
					self._print_indent("NO DOMINANTE!")			
		return 0
	
	
	#
	# RUTINA PRINCIPAL
	# Convierte la red de petri en un grafo de alcanzabilidad de forma recursiva
	#
	def convert ( self, pn_atom3, aCurrentState=None, oParentNode=None, rg_atom3 = None, iPos = 1, transition = None) :
		
		# Obtener la Red de Petri
		pn = pn_atom3.ASGroot
		
		# Obtener el grafo de alcanzabilidad en su estado actual
		# si no existe lo creamos
		if rg_atom3 == None :
			# Nueva ventana de ATOM3
			rg_atom3 = ATOM3(Tk(), 'ReachGraph', 0, 1)
			aCurrentState = self.pn_getCurrentState( pn )
		else :
			self.iRec = self.iRec + 1
		rg = rg_atom3.ASGroot
	
		self._print_indent("---------------------")
		if aCurrentState != None :
		  self._print_indent("Estado=" + self._list2str(self._sort_dict(aCurrentState)))
			
		# Obtener la lista de nuevos estados candidatos, a partir
		# de las transiciones activas
	        aNewStates = self.pn_getCandidates ( pn, aCurrentState )

		# Añadir el nodo actual	
		oCurrentNode = self.rg_addNode( rg_atom3, aCurrentState, oParentNode, iPos, transition )
	
		# Para cada uno de estos estados, avanzamos la red y llamada recursiva
		oSameOlder = 0
		iChildPos=0
		for s, t in aNewStates:
			iChildPos=iChildPos+1
			self._print_indent("Estudiando=" +  self._list2str(self._sort_dict(s)) + " desde " +  self._list2str(self._sort_dict(aCurrentState)))

			# Si el estado ya lo hemos visto, rompemos la recursion
			oSameOlder = self.rg_findState ( rg, s )
			if oSameOlder == 0 :
				# Convertimos la red a partir del nodo actual
	        		oSameOlder = self.convert( pn_atom3, s, oCurrentNode, rg_atom3, iChildPos, t)
			# Si hemos encontrado el mismo estado en el grafo
			# deberemos enlazar al padre de este con el recien
			# encontrado
			else :
				self.rg_connectPair( rg_atom3, oCurrentNode, oSameOlder, t )

		self.iRec = self.iRec-1
		return oSameOlder

